#include "ystring.h"
#include "log.h"
#include <process.h>
#include <io.h>
#include <sys/locking.h>

int file_isregular(int fd);
static int LockingX(int fd,int type,int bytes);
int lock_dbg = 0;
int lock_ext = 1;

static int Locking(int fd,int type,int bytes)
{ 	int off,rcode;

	if( lock_ext ){
		return LockingX(fd,type,bytes);
	}

	rcode = -1;
	if( 0 <= (off = tell(fd)) )
	if( 0 <= lseek(fd,0,0) ){
		rcode = locking(fd,type,bytes);
		lseek(fd,off,0);
	}
	if( rcode == -1 && !file_isregular(fd) ){
		/* treat as success, 
		 * mainly for lock_TIMEOUT not to wait non-lockable file ... ?
		 */
		rcode = 0;
	}
	return rcode;
}

int SHlockB(int fd){	return	Locking(fd,_LK_RLCK,1); }
int SHlockNB(int fd){	return	Locking(fd,_LK_NBRLCK,1); }
int EXlockB(int fd){	return	Locking(fd,_LK_LOCK,1); }
int EXlockNB(int fd){	return	Locking(fd,_LK_NBLCK,1); }
int UNlock(int fd){	return	Locking(fd,_LK_UNLCK,1); }


int porting_dbg(const char *fmt,...);
static const char *lksym(int type){
	switch(type){
		case _LK_RLCK:  return "RLCK";   break;
		case _LK_NBRLCK:return "NBRLCK"; break;
		case _LK_LOCK:  return "LOCK";   break;
		case _LK_NBLCK: return "NBLCK";  break;
		case _LK_UNLCK: return "UNLCK";  break;
	}
	return "?";
}
static void lklog(int fd,int type,int bytes,const char *fmt,...){
	VARGS(16,fmt);
	CStr(msg,1024);
	const char *st;
	int xerr;

	if( !file_isregular(fd) ) /* not the tty for LOGFILE */
		return;

	xerr = GetLastError();
	sprintf(msg,"Locking(%d,%-6s,%d)[xerr=%d] ",fd,lksym(type),bytes);
	Xsprintf(TVStr(msg),fmt,VA8);
	porting_dbg("%s",msg);
}

static int dolock1(int fd,int type,FileSize lkoff,FileSize bytes){
	HANDLE oh;
	int offh,offl,lenh,lenl;
	int rcode;
	int xerr;

	oh = (HANDLE)_get_osfhandle(fd);
	offh = lkoff >> 32;
	offl = lkoff & 0xFFFFFFFF;
	lenh = bytes >> 32;
	lenl = bytes & 0xFFFFFFFF;
	SetLastError(0);

	if( type == _LK_UNLCK ){
		if( UnlockFile(oh,offl,offh,lenl,lenh) )
			rcode = 0;
		else	rcode = -1;
	}else{
		if( LockFile(oh,offl,offh,lenl,lenh) )
			rcode = 0;
		else	rcode = -1;
	}

	if( lock_dbg || lDEBUGLOCK() ){
		xerr = GetLastError();
		porting_dbg("-- LK [%d/%d] %-6s %llX %llX = %d xerr=%d",
			fd,oh,lksym(type),lkoff,bytes,rcode,xerr);
	}
	return rcode;
}
static int dolock(int fd,int type,FileSize lkoff,FileSize bytes){
	int ti;
	int rcode;

	for( ti = 0; ti < 10; ti++ ){
		rcode = dolock1(fd,type,lkoff,bytes);
		if( rcode == 0 )
			break;
		if( type == _LK_NBLCK || type == _LK_NBRLCK )
			break;
		if( type == _LK_UNLCK )
			break;
		if( lock_dbg || lDEBUGLOCK() ){
			porting_dbg("-- lock retry [%d] %d\n",ti,type);
		}
		sleep(1);
	}
	return rcode;
}
static int LockingX(int fd,int type,int xbytes){
	int shared;
 	int rcode;
	int pid;
	int bytes = 0;
	FileSize poff;
	FileSize loff;

	shared = (type == _LK_RLCK || type == _LK_NBRLCK);
	rcode = -1;
	if( 0 <= tell(fd) ){
		pid = _getpid() & 0xFFFF;
		if( shared ){
			poff = pid;
			bytes = 1;
		}else{
			poff = 0;
			bytes = 0x10000;
		}

		loff = 0x10 * 0x100000000;
		poff += loff;

		rcode = dolock(fd,type,poff,bytes);
		if( rcode != 0 && type == _LK_UNLCK ){
			if( shared ){
				poff = 0;
				bytes = 0x10000;
			}else{
				poff = pid;
				bytes = 1;
			}
			poff += loff;
			rcode = dolock(fd,type,poff,bytes);
		}
		if( type != _LK_UNLCK )
		if( type != _LK_NBRLCK && type != _LK_NBLCK ){
			if( rcode != 0 || lock_dbg || lDEBUGLOCK() ){
				lklog(fd,type,bytes,"=%d",rcode);
			}
		}
		errno = 0;
	}else{
		lklog(fd,type,bytes,"tell error");
		errno = 0;
	}
	if( rcode == -1 && !file_isregular(fd) ){
		/* treat as success, 
		 * mainly for lock_TIMEOUT not to wait non-lockable file ... ?
		 */
		rcode = 0;
	}
	return rcode;
}
